/*
 * laser3d - 3D visualization of laser data acquired by robot; enables walk through
 *
 * Filename    : data.c
 * Author      : Michael Kaess
 * Initial     : 02/15/2001
 * Last Change : 12/27/2001
 *
 */


#include <stdlib.h>
#include <stdio.h>

#include "vertex.h"
#include "laser3d.h"

#include "data.h"


/* maximum line length of header of raw pgm file */
#define MAX_LINE_LEN 40



/***** global variables *****/


/* original data read from file */
TRawScan *RawScan;
/* number of records allocated */
int NumRecs = 0;

/* default resolution: mm */
float ScaleLaser = SCALE_LASER_MM;
unsigned int LaserMaxValue = LASER_MAX_VALUE_MM;

/* 16 (default) or 8 bit mode for writing pgm files */
int PgmMode = PGM_16_BIT;

/* only for 360 degree round scans: remove unnecessary data */
int Correct360Mode = FALSE;

/* correct scans in 3D coordinates, see graphics.c; */
/* will be filled by this module                    */
extern Vertex **Scan;

/* straight or round scan, from graphics; */
/* needed for transformation to 3D        */
extern int ScanMode;

/* print information about processed data? */
extern int Verbose;

/* initial and current viewing direction, see graphics.c; */
/* will be reversed here if robot moved backwards         */
extern float StartEyeAngle[];
extern float EyeAngle[];


/***** local functions *****/


#ifdef SGI

/* reverse byte order for float */
void convFloat(float *f) { //SGI only
  float x;
  unsigned char *px = (unsigned char *)(f);
  unsigned char *re = (unsigned char *)(&x);
  *re = *(px+3);
  *(re+1) = *(px+2);
  *(re+2) = *(px+1);
  *(re+3) = *px;
  
  *f = x;
}


/* reverse byte order for integer */
void convInt(unsigned int *i) { //SGI only
  unsigned int x;
  unsigned char *px = (unsigned char *)(i);
  unsigned char *re = (unsigned char *)(&x);
  *re = *(px+3);
  *(re+1) = *(px+2);
  *(re+2) = *(px+1);
  *(re+3) = *px;
  
  *i = x;
}

#endif


/* read in laser data from file created by laserfit (hserver/MissionLab) */
int readFile(char *fname) {
  FILE *inFile;
  int num;
  TRawScan temp;
  
  inFile = fopen(fname, "rb");
  if (inFile == NULL) {
    
    printf("ERROR: Could not open input file %s\n", fname);
    exit3d(1);
    
  } else {

    /* count number of records in file */
    NumRecs=0;
    while ((!feof(inFile))) {
      fread(&temp, sizeof(TRawScan), 1, inFile);
      NumRecs++;
    }
    if (Verbose == TRUE) {
      printf("--Number of scans from file: %i\n", NumRecs);
    }

    /* finally allocate memory for the raw data */
    /* now that the amount of data is known     */
    RawScan = (PRawScan)malloc(NumRecs*sizeof(TRawScan));
    if (RawScan == NULL) {
      printf("ERROR: Couldn't allocate memory\n");
      exit3d(1);
    }

    /* reset pointer to start of file */
    fseek(inFile, 0, SEEK_SET);
    /* read data */
    num = 0;
    while ((!feof(inFile))) {
      fread(&RawScan[num], sizeof(TRawScan), 1, inFile);
#ifdef SGI
      for (int ii=0; ii<NUM_SCANS; ii++) {
        convInt(&RawScan[num].dist[ii]);
      }
      convFloat(&RawScan[num].x);
      convFloat(&RawScan[num].y);
      convFloat(&RawScan[num].t);
#endif
#ifdef DATAFILE_USES_DEGREES
      RawScan[num].t=RawScan[num].t*(2*M_PI/360.0); /* convert to radians */
#endif
      num++;
    }
    fclose(inFile);
  }

  return num;
}


/* hserver sometimes doesn't update the pose of the robot   */
/* for a few steps, therefore interpolation of the position */
/* and the angle are necessary                              */
int interpolPos(int start, int end) {
  int num, num2, c, fac;
  float dx, dy, dt;

  /* go through all valid scans */
  num = start;
  do {
    /* search for a set of scans with equal pose */
    num2 = num + 1;
    while ((RawScan[num].x == RawScan[num2].x) && (RawScan[num].y == RawScan[num2].y)
           && (RawScan[num].t == RawScan[num2].t) && (num2 <= end)) {
      num2++;
    }
    /* at the end of the data set? -> cannot be corrected reliably, cut them off */
    if ((RawScan[num].x == RawScan[num2].x) && (RawScan[num].y == RawScan[num2].y)
        && (RawScan[num].t == RawScan[num2].t) && (num2 > end)) {
      end = num;
    } else {
      /* do interpolation over scans with equal pose */
      dx = (RawScan[num2].x-RawScan[num].x)/(num2-num);
      dy = (RawScan[num2].y-RawScan[num].y)/(num2-num);
      if ((RawScan[num2].t-RawScan[num].t)>300.0) {
        dt = (RawScan[num2].t-RawScan[num].t-360)/(num2-num); /* 360 degrees! */
      } else if ((RawScan[num2].t-RawScan[num].t)<-300.0) {
        dt = (RawScan[num2].t-RawScan[num].t+360)/(num2-num); /* -360 degrees! */
      } else {
        dt = (RawScan[num2].t-RawScan[num].t)/(num2-num);
      }
      fac = 1;
      for (c=num+1; c<num2; c++) {
        RawScan[c].x += fac*dx;
        RawScan[c].y += fac*dy;
        RawScan[c].t += fac*dt;
        fac++;
      }
      num = num2;
    }
  } while (num < end);

  return end;
}


/* change position of the robot so that it starts at the origin; */
/* makes displaying the data easier */
void nullOffset(int start, int end) {
  int num;
  float x, y, x2, y2, t;
  int dir = 1; /* robot moved forward */
  float diff;

  x = RawScan[start].x;
  y = RawScan[start].y;
  /* test if robot moved backwards */
  t = atan2(RawScan[end].y-y, RawScan[end].x-x);
  diff = fabs(t-RawScan[start].t);
  if (diff > 90*(M_PI/180.0)) {
    dir = -1; /* robot moved backwards */
  }
  /* TODO: this assumes a STRAIGHT trajectory! */
  /* should be easy to generalize, but may result in    */
  /* wrong surfaces in the model if robot turns fast... */
  for (num=start; num<=end; num++) {
    x2 = RawScan[num].x - x;
    y2 = RawScan[num].y - y;
    RawScan[num].y = dir * sqrt(x2*x2 + y2*y2);
    RawScan[num].x = 0;
    RawScan[num].t = 0;
  }
  if (Verbose == TRUE) {
    printf("--Average movement of the robot between scans: %fm.\n", RawScan[end].y/(end-start));
  }

  /* rotate start position of user if robot moved backwards */
  if (dir == -1) {
    StartEyeAngle[0] += 180;
    EyeAngle[0] += 180;
    if (Verbose == TRUE) {
      printf("--Robot moved backwards, correcting orientation.\n");
    }
  }
}


/* remove invalid and useless scans from the beginning and end */
/* of the gathered data; useless means repeated records from   */
/* the same pose of the robot while it didn't move or records  */
/* with invalid pose information (may happen in the beginning) */
void getStartAndEnd(int *start, int *end) {
  int pos;
  float x, y, t;
  
  pos = *start;
  /* skip all entries at the beginning which have invalid values */
  while (((RawScan[pos].x == 0) && (RawScan[pos].y == 0) && (RawScan[pos].t == 0)) ||
         (isnan(RawScan[pos].x)) || (isnan(RawScan[pos].y)) || (isnan(RawScan[pos].t))) {
    pos++;
  }
  /* skip all entries at the beginning which are equal */
  /* (the robot didn't move yet)                       */
  x = RawScan[pos].x;
  y = RawScan[pos].y;
  t = RawScan[pos].t;
  while ((RawScan[pos].x == x) && (RawScan[pos].y == y) && (RawScan[pos].t == t)) {
    pos++;
  }
  pos--;
  *start = pos;

  pos = *end;
  /* remove all entries at the end which are equal (robot didn't move) */
  x = RawScan[pos].x;
  y = RawScan[pos].y;
  t = RawScan[pos].t;
  while ((RawScan[pos].x == x) && (RawScan[pos].y == y) && (RawScan[pos].t == t)) {
    pos--;
  }
  pos++;

  if (ScanMode == STRAIGHT_SCAN) { /*??? shouldn't be necessary anymore */ /*TODO*/
    pos-=2;
  }
  
  *end = pos;
}


/* cuts unnecessary scans for 360 degree round scan mode */
void correctRoundScan(int *s, int *e) {
  int start, end;
  
  /* TODO: this is not a general implementation, */
  /* has to be changed depending on experiment   */
  
  start = *s;
  end = *e;
  while (RawScan[start].t < 100.0*(M_PI/180.0)) { /* start with t around 360 degrees */
    start++;
  }
  while ((RawScan[end].t <= RawScan[start].t) && (RawScan[end].t > 100.0*(M_PI/180.0))) {
    end--; /* end without overlap */
  }
  if (Verbose == TRUE) {
    printf("--Corrected for 360 degree roundscan: Displaying data from scan %i to %i.\n", start, end);
  }

  /* return the new range of "good" data */
  *s = start;
  *e = end;
}


/* only for round scan:                     */
/* conversion from raw data to Cartesian 3D */
void rawTo3dRound(int start, int end) {
  int num, num3d, c;
  float  alpha, ainc, x, y, t;
  int inv = 0; /* number of invalid data points, i.e. too far away */
  
  ainc = M_PI/NUM_SCANS;

  num3d = 0;
  /* process all valid scans */
  for (num=start; num<=end; num++) {

    alpha = -M_PI/2.0 + M_PI/180.0*SCAN_TILT_ROUND; /* correction for tilted laser scanner */
    x = RawScan[num].x;
    y = RawScan[num].y;
    t = RawScan[num].t;

    for (c=0; c<NUM_SCANS; c++) {
      /* check if point is within valid distance: */
      /* laser returns maximum value if no object is detected within its range */
      if (RawScan[num].dist[c] >= LaserMaxValue) {
          Scan[num3d][c+1].x = INVALID;
          alpha += ainc;
        inv++;
      } else {
        Scan[num3d][c+1] = Vertex((x + cos(t)*(SCAN_RADIUS_ROUND+RawScan[num].dist[c]*cos(alpha)))*ScaleLaser*SCALE_X + OFFSET_X,
                                  (y + sin(t)*(SCAN_RADIUS_ROUND+RawScan[num].dist[c]*cos(alpha)))*ScaleLaser*SCALE_Y + OFFSET_Y,
                                  (SCAN_HEIGHT_ROUND + (RawScan[num].dist[c]*sin(alpha))*ScaleLaser)*SCALE_Z + OFFSET_Z
                                  );
        alpha += ainc;
      }
    }
    num3d++;
  }
  if (Verbose == TRUE) {
    printf("--%i invalid points\n", inv);
  }
}


/* only for straight movement of the robot: */
/* conversion from raw data to Cartesian 3D */
void rawTo3dStraight(int start, int end) {
  int num, num3d, c;
  float  alpha, ainc, theta, x, y;
  int inv = 0; /* number of invalid data points, i.e. too far away */
  unsigned int raw;
  
  ainc = -M_PI/NUM_SCANS;

  num3d = 0;
  /* process all valid scans */
  for (num=start; num<=end; num++) {
    theta = RawScan[num].t;
    x = RawScan[num].x;
    y = RawScan[num].y;
    alpha = M_PI;
    for (c=0; c<NUM_SCANS; c++) {
      //raw = (RawScan[num].dist[c]/16)*16; /* lower resolution; for test purposes */
      raw = RawScan[num].dist[c];
      /* check if point is within valid distance: */
      /* laser returns maximum value if no object is detected within its range */
      if (raw >= LaserMaxValue) {
          Scan[num3d][c+1] = Vertex(INVALID, INVALID, INVALID);
          alpha += ainc;
        inv++;
      } else {
        Scan[num3d][c+1] = Vertex((x + raw*ScaleLaser*cos(alpha)*cos(theta))*SCALE_X + OFFSET_X,
                                  (y + raw*ScaleLaser*cos(alpha)*sin(theta))*SCALE_Y + OFFSET_Y,
                                  (SCAN_HEIGHT_STRAIGHT + raw*ScaleLaser*sin(alpha))*SCALE_Z + OFFSET_Z
                                  );
        alpha += ainc;
      }
    }
    /* project first and last measurement down to floor to make model more */
    /* realistic; necessary since laser scanner doesn't measure below the height */
    /* it is mounted at, i.e. the height of the robot */
    Scan[num3d][0] = Vertex(Scan[num3d][1].x, Scan[num3d][1].y, 0);
    Scan[num3d][NUM_SCANS+1] = Vertex(Scan[num3d][NUM_SCANS].x, Scan[num3d][NUM_SCANS].y, 0);
    num3d++;
  }
  if (Verbose == TRUE) {
    printf("--%i invalid points\n", inv);
  }
}



/***** global functions *****/


/* convert the raw data (radial distances) to Cartesian 3D coordinates */
void datRawTo3d(int start, int end) {
  /* there are no big differences between the two versions */
  /* of rawTo3d, but combining them and using "if"s is not */
  /* a good idea due to the big loops and the resulting    */
  /* slow execution */
  if (ScanMode == ROUND_SCAN) {
    rawTo3dRound(start, end);
  } else {
    rawTo3dStraight(start, end);
  }
}


/* interpolates over invalid measurements;                        */
/* when the laser hits surfaces almost parallel to the laser beam */
/* like the edge of a picture frame, it sometimes cannot detect   */
/* the object at all and returns maximum distance; these errors   */
/* can simply be corrected by interpolation;                      */
/* note that this will not interpolate over invalid areas higher  */
/* than one pixel AND wider than one pixel; such areas typically  */
/* are really empty within the range of the laser scanner         */
void datInterpolateData(int start, int end) {
  int num;
  int c;
  int ipol1, ipol2;

  ipol1 = 0;
  ipol2 = 0;
  /* do interpolation along a 180 degree scan (along the 361 measurements) */
  for (num=start; num<=end; num++) {
    for (c=0; c<NUM_SCANS; c++) {
      /* if invalid, check direct neighbors; if they are valid, use them to interpolate */
      if (RawScan[num].dist[c] >= LaserMaxValue) {
        if ((RawScan[num+1].dist[c] < LaserMaxValue) && (RawScan[num-1].dist[c] < LaserMaxValue)) {
          RawScan[num].dist[c] = (RawScan[num+1].dist[c] + RawScan[num-1].dist[c])/2;
          ipol1++;
        }
      }
    }
  }
  /* do interpolation between different 180 degree scans (along scans with the same angle) */
  for (c=0; c<NUM_SCANS; c++) {
    for (num=start; num<=end; num++) {
      /* if invalid, check direct neighbors; if they are valid, use them to interpolate */
      if (RawScan[num].dist[c] >= LaserMaxValue) {
        if ((RawScan[num].dist[c+1] < LaserMaxValue) && (RawScan[num].dist[c-1] < LaserMaxValue)) {
          RawScan[num].dist[c] = (RawScan[num].dist[c+1] + RawScan[num].dist[c-1])/2;
          ipol2++;
        }
      }
    }
  }
  if (Verbose == TRUE) {
    printf("--Interpolated over %i and %i errors in x and y direction\n", ipol1, ipol2);
  }
}


/* write raw scan data to 8 or 16 bit raw pgm format file;      */
/* 16 bit mode writes the complete data, that is all 13 bits we */
/* get from the laser scanner;                                  */
/* 8 bit mode only makes visualization of the images easier     */
/* since most graphic programs don't support the 16 bit version */
void datWriteRawPgmFile(char *fname, int start, int end) {
  FILE *fout;
  int pos, num;

  fout = fopen(fname, "w");
  if (fout == NULL) {
    printf("ERROR: Cannot open file %s\n", fname);
    exit3d(1);
  } else {
    if (PgmMode == PGM_16_BIT) {
      /* 16 bit PGM, 8192 used values */
      fprintf(fout, "P5\n361 %i\n8191\n", end-start);
      for (num=start; num<end; num++) {
        for (pos=0; pos<NUM_SCANS; pos++) {
          fprintf(fout, "%c%c", (unsigned char)(RawScan[num].dist[pos]%256), (unsigned char)(RawScan[num].dist[pos]/256));
        }
      }
      if (Verbose == TRUE) {
        printf("--Raw pgm format file %s written (16 bit format).\n", fname);
      }
    } else {
      /* 8 bit PGM, all 256 values used */
      fprintf(fout, "P5\n361 %i\n255\n", end-start);
      for (num=start; num<end; num++) {
        for (pos=0; pos<NUM_SCANS; pos++) {
          fprintf(fout, "%c", (unsigned char)(RawScan[num].dist[pos]/16));
        }
      }
      if (Verbose == TRUE) {
        printf("--Raw pgm format file %s written (8 bit format).\n", fname);
      }
    }
    fclose(fout);
  }
}


/* read raw scan data from 8 or 16 bit raw pgm format;                */
/* a modified (for example smoothened) file can be displayed that way */
int datReadRawPgmFile(char *fname, int *s) {
  FILE *fin;
  unsigned char str[MAX_LINE_LEN];
  int pos, num;
  unsigned char c1, c2;
  int x, len; /* image dimensions */
  int depth;
  int start, end;

  start = *s;
  fin = fopen(fname, "r");
  if (fin == NULL) {
    printf("ERROR: Cannot open file %s\n", fname);
    exit3d(1);
  } else {
    /* process header */
    /* magic number for PGM */
    fgets((char *)str, MAX_LINE_LEN, fin);
    if ((str[0]!='P') && (str[1]!='5')) {
      printf("ERROR: %s not in raw pgm format\n", fname);
      exit3d(1);
    }
    /* image dimensions */
    fscanf(fin, "%i %i%[^\n]\n", &x, &len, str);
    if (x!=361) {
      printf("ERROR: image in %s has wrong dimensions\n", fname);
      exit3d(1);
    }
    /* color depth */
    fgets((char *)str, MAX_LINE_LEN, fin);
    fscanf(fin, "%i[^\n]\n", &depth);
    fscanf(fin, "%c", str); /* read last \n */
    if ((depth!=8191) && (depth!=255)) {
      printf("ERROR: image format in %s uses %i different values instead of 8192 or 256\n", fname, depth+1);
      exit3d(1);
    }

    if (RawScan == NULL) {
      RawScan = (PRawScan)malloc(len*sizeof(TRawScan));
    }
    if (RawScan == NULL) {
      printf("ERROR: Couldn't allocate memory\n");
      exit3d(1);
    }
    end = start + len;
    if (depth == 8191) {
      /* read 16 bit raw data */
      for (num=start; num<end; num++) {
        for (pos=0; pos<NUM_SCANS; pos++) {
          fscanf(fin, "%c%c", &c2, &c1);
          RawScan[num].dist[pos] = c1*256 + c2;
        }
      }
    } else {
      /* read 8 bit raw data */
      for (num=start; num<end; num++) {
        for (pos=0; pos<NUM_SCANS; pos++) {
          fscanf(fin, "%c", &c1);
          RawScan[num].dist[pos] = c1*16;
        }
      }
      printf("WARNING: Data is in 8 bit raw pgm format -> reduced depth resolution\n");
    }
    fclose(fin);
    if (Verbose == TRUE) {
      printf("--Data from raw pgm format file %s read, contains %i scans\n", fname, len);
    }
  }

  *s = start;
  
  return end;
}


/* writes the cleaned and corrected scan data back to a file in  */
/* the laser data format that laserfit (hserver/MissionLab) uses */
void datWriteCleanedFile(char *fname, int start, int end) {
  FILE *outFile;
  int i;
  float tmp;

  outFile = fopen(fname, "wb");
  if (outFile != NULL) {
    /* write only valid scans */
    for (i=start; i<=end; i++) {
#ifdef DATAFILE_USES_DEGREES
      tmp = RawScan[i].t;
      RawScan[i].t=RawScan[i].t/(2*M_PI/360.0); /* convert back to degrees */
#endif
      fwrite(&RawScan[i], sizeof(TRawScan), 1, outFile);
#ifdef DATAFILE_USES_DEGREES
      RawScan[i].t = tmp; /* restore radians */
#endif
    }
    fclose(outFile);
  }
}
 

/* read in data from file; get valid range of useful data */
/* and do further corrections                             */
void datGetRaw3dData(char *fname, int *s, int *e) {
  int num;
  int start, end;
  
  num = readFile(fname);

  start = 0;
  end = num-1;
  
  /* clip unusable data */
  getStartAndEnd(&start, &end);
  /* repair data, due to problem with hserver */
  end = interpolPos(start, end);
  
  if (Verbose == TRUE) {
    printf("--Displaying data from scan %i to %i.\n", start, end);
  }

  if (ScanMode == STRAIGHT_SCAN) {
    nullOffset(start, end); /* only for straight hallway */
  } else {
    if (Correct360Mode == TRUE) {
      correctRoundScan(&start, &end); /* only for roundscan */
    }
  }

  /* return frame with valid scans */
  *s = start;
  *e = end;
}


/* changes between 8 and 16 bit mode for writing pgm files */
void datSetPgmMode(int mode) {
  PgmMode = mode;
}


/* do additional correction of data: for 360 degree round scans only; */
/* remove all but the data of one complete 360 degree turn            */
void datSet360Mode(void) {
  Correct360Mode = TRUE;
}


/* changes the mode of the laser scanner between */
/* centimeter and millimeter resolution          */
void datSetResolution(int res) {
  if (res == RESOLUTION_CM) {
    ScaleLaser = SCALE_LASER_CM;
    LaserMaxValue = LASER_MAX_VALUE_CM;
  } else {
    ScaleLaser = SCALE_LASER_MM;
    LaserMaxValue = LASER_MAX_VALUE_MM;
  }
}


/* free previously allocated memory */
void datFree(void) {
  if (RawScan != NULL) {
    free(RawScan);
  }
}
