/*
 * laser3d - 3D visualization of laser data acquired by robot; enables walk through
 *
 * Filename    : graphics.c
 * Author      : Michael Kaess
 * Initial     : 02/15/2001
 * Last Change : 12/27/2001
 *
 * OpenGL code based on CS6490 project, Ralf Habel/Michael Kaess, Fall 2000
 * view parameters and vertex class based on bsp code, Mark Carlson
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>

#include "vertex.h"
#include "laser3d.h"

#include "graphics.h"



/***** global variables *****/


/* initial window size */
int WindowWidth  = WINDOW_WIDTH;
int WindowHeight = WINDOW_HEIGHT;

/* initial window position */
int WindowPosX = WINDOW_POS_X;
int WindowPosY = WINDOW_POS_Y;

/* initial eye point in the model:           */
/* (0,0,h) corresponds to the start point of */
/* the scan with eye height of h in meters   */
Vertex StartPosition(0.0f, 0.0f, EYE_HEIGHT);

/* initial viewing direction (x,z in degrees): */
/* (0,-90) to start direction of the scan      */
float StartEyeAngle[2] = {0.0f, -90.0f};

/* internally used for navigating the model */
Vertex MovePosition = StartPosition;
Vertex EyePosition = MovePosition;
float EyeAngle[2] = {StartEyeAngle[0], StartEyeAngle[1]};

static const float ClippingFar = 100.0f;
static const float ClippingNear = (float)0.500000;
float WindowFieldOfView = 60.0f;

/* correct scans in 3D coordinates */
Vertex **Scan;
/* normals of the 2 triangles defined by each scan point */
Vertex **Normal0, **Normal1;

/* how to display the model and the artificial floor */
int Mode      = MODE_SHADED;
int FloorMode = MODE_FLOOR_CHECKER;
/* has model changed? if yes, redraw is not enough */
/* GlList has to be recalculated */
int ModeChange = TRUE;

/* number of correct scans */
int Num3d;

/* straight or round scan */
int ScanMode = STRAIGHT_SCAN; /* default mode: straight scan */

/* GL feature to avoid repeated passing of all 3d model */
/* information when changing the viewpoint              */
GLuint GlList;



/***** local functions *****/


/* makes sure the user is looking in the right direction */
void transformWorld(void) {

  /* make z up instead of y up */
  /* up down */
  glRotatef(EyeAngle[1], 1.0f, 0.0f, 0.0f);

  /* left right */
  glRotatef(EyeAngle[0], 0.0f, 0.0f, 1.0f);

  glTranslatef(-EyePosition.x, -EyePosition.y, -EyePosition.z);
}


/* passes data to OpenGL, format depends on mode */
void drawScene(void) {
  int num, pos;
  int newStart; /* boolean, true if strip starts, used for finding the right normals */
  int x, y;
  int start, end;

  /* show user that program is busy */
  glutSetCursor(GLUT_CURSOR_WAIT);

  /* have to transfer all model data to OpenGL only  */
  /* when model itself changed; previously generated */
  /* GlList can be used when only viewpoint changes  */
  if (ModeChange == TRUE) {
    ModeChange = FALSE;
    
    glNewList(GlList, GL_COMPILE_AND_EXECUTE);

    glNormal3f(0, 0, 1);
    
    /* generate floor */
    if (FloorMode == MODE_FLOOR_GREEN) {
      /* single color floor */
      glBegin(GL_QUADS);
      glColor3f(0.0, 1.0, 0.0); /* green */
      glVertex3f(-50.0, -50.0, 0.0);
      glVertex3f(50.0, -50.0, 0.0);
      glVertex3f(50.0, 50.0, 0.0);
      glVertex3f(-50.0, 50.0, 0.0);
      glEnd();
    } else if (FloorMode == MODE_FLOOR_CHECKER) {
      /* checkerboard, size given by the length of the sides of */
      /* each quad: CB_QUAD_SIZE, and the number of quads in    */
      /* each direction: CB_GRID_NUM                            */
      glBegin(GL_QUADS);
      for (x=-CB_GRID_NUM; x<CB_GRID_NUM; x++) {
        for (y=-CB_GRID_NUM; y<CB_GRID_NUM; y++) {
          if (((x+CB_GRID_NUM)%2)==((y+CB_GRID_NUM)%2)) {
            glColor3f(0.0, 0.0, 0.0); /* black quad */
          } else {
            glColor3f(1.0, 1.0, 1.0); /* white quad */
          }
          glVertex3f(CB_QUAD_SIZE*x, CB_QUAD_SIZE*y, 0.0);
          glVertex3f(CB_QUAD_SIZE*(x+1.0), CB_QUAD_SIZE*y, 0.0);
          glVertex3f(CB_QUAD_SIZE*(x+1.0), CB_QUAD_SIZE*(y+1.0), 0.0);
          glVertex3f(CB_QUAD_SIZE*x, CB_QUAD_SIZE*(y+1.0), 0.0);
        }
      }
      glEnd();
    }

    /* generate mesh from scanned data */

    if ((Mode == MODE_SHADED) || (Mode == MODE_TRI)) {
      /* shaded triangles / random color triangles */

      /* for each strip */
      for (num=0; num<Num3d-1; num++) {
        glColor3f(0.7, 0.7, 0.7);
        glBegin(GL_TRIANGLE_STRIP);
        newStart = TRUE; /* normals for shading different for first triangles within a strip */
        if (ScanMode == ROUND_SCAN) {
          start = 1;
          end = NUM_SCANS+1;
        } else {
          /* for straight scan: use projection of start and end points to floor */
          start = 0;
          end = NUM_SCANS+2;
        }
        /* for each scan point within one 180 degree scan */
        for (pos=start; pos<end; pos++) {
          /* don't process invalid points */
          if ((Scan[num][pos].x != INVALID) && (Scan[num+1][pos].x != INVALID)) {
            if (Mode == MODE_TRI) {
              glColor3f(rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX);
            }
            if (newStart == TRUE) {
              glNormal3f(Normal0[num][pos].x, Normal0[num][pos].y, Normal0[num][pos].z);
            } else {
              glNormal3f(Normal0[num][pos-1].x, Normal0[num][pos-1].y, Normal0[num][pos-1].z);
            }
            glVertex3f(Scan[num][pos].x, Scan[num][pos].y, Scan[num][pos].z);
            if (Mode == MODE_TRI) {
              glColor3f(rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX);
            }
            if (newStart == TRUE) {
              glNormal3f(Normal1[num][pos].x, Normal1[num][pos].y, Normal1[num][pos].z);
              newStart = FALSE;
            } else {
              glNormal3f(Normal1[num][pos-1].x, Normal1[num][pos-1].y, Normal1[num][pos-1].z);
            }
            glVertex3f(Scan[num+1][pos].x, Scan[num+1][pos].y, Scan[num+1][pos].z);
          } else {
            glEnd();
            newStart = TRUE; /* new strip may start; do normals appropriately */
            glBegin(GL_TRIANGLE_STRIP);
          }
        }
        glEnd();
      }

    } else if ((Mode == MODE_STRIP) || (Mode == MODE_QUADS)) {
      /* random color quad strips or quads */
      
      for (num=0; num<Num3d-1; num++) {
        glBegin(GL_QUAD_STRIP);
        if (Mode == MODE_STRIP) {
          glColor3f(rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX);
        }
        if (ScanMode == ROUND_SCAN) {
          start = 1;
          end = NUM_SCANS+1;
        } else {
          /* for straight scan: use projection of start and end points to floor */
          start = 0;
          end = NUM_SCANS+2;
        }
        for (pos=start; pos<end; pos++) {
          /* don't process invalid points */
          if ((Scan[num][pos].x != INVALID) && (Scan[num+1][pos].x != INVALID)) {
            if (Mode == MODE_QUADS) {
              glColor3f(rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX, rand()*1.0/RAND_MAX);
            }
            glVertex3f(Scan[num][pos].x, Scan[num][pos].y, Scan[num][pos].z);
            glVertex3f(Scan[num+1][pos].x, Scan[num+1][pos].y, Scan[num+1][pos].z);
          } else {
            glEnd();
            glBegin(GL_QUAD_STRIP);
          }
        }
        glEnd();
      }
    
    }
      
    glEndList();

  }

  /* background black */
  glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();

	transformWorld();

  glCallList(GlList);

  /* force update, and swap buffers so that the transition is smooth */
  glFlush();
  glutSwapBuffers();

  /* show user that program is ready now */
  glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
}


/* sets up view frustum, called when graphics window is created */
void changeFrustum(void) {
  double frustumHeight;
  double frustumWidth;

  /* set up a projection matrix */
  glMatrixMode(GL_PROJECTION);
  
  glLoadIdentity();

  /* set up the Field of View */
  frustumWidth = ClippingNear*tan(0.50*(WindowFieldOfView)*M_PI/180.0);
  frustumHeight = ((double)WindowHeight / (double)WindowWidth) * frustumWidth;
  glFrustum(-frustumWidth, frustumWidth, -frustumHeight, frustumHeight,
            ClippingNear, ClippingFar);

  glViewport(0, 0, WindowWidth, WindowHeight);

  /* create nice light (parameters are somewhat random...) */
  GLfloat light_ambient[]={0.4, 0.4, 0.4, 0.0};
  GLfloat light_diffuse[]={0.5, 0.5, 0.5, 0.0};
  GLfloat light_specular[]={0.6, 0.6, 0.6, 0.0};

  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);

  /* put light source right above the observer */
  GLfloat light_pos[3]={0, 0, EYE_HEIGHT + 0.8};
  glLightfv(GL_LIGHT0, GL_POSITION, light_pos);

  //  glFrontFace(GL_CW);
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
  //  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  //  glPolygonMode(GL_FRONT, GL_FILL);

  glShadeModel(GL_FLAT);
  // glShadeModel(GL_SMOOTH);

	glEnable(GL_DEPTH_TEST);

  /* background black */
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* define the modelview transformation */
  glMatrixMode(GL_MODELVIEW);
}


/* processes keyboard events */
void keyboard(unsigned char key, int x, int y) {
	double zAngle;
	double o, a;

  /* calculate increment step depending on current orientation */
  zAngle = EyeAngle[0] * M_PI/180.0;
  o = MOVE_STEP * sin(zAngle); /* x axis is right */
  a = MOVE_STEP * cos(zAngle); /* y axis is forward */

  switch (key) {
    case EVENT_EXIT:
    case EVENT_EXIT2:
      exit3d(0);
      break;
    case EVENT_RESET:
      EyeAngle[0] = StartEyeAngle[0];
      EyeAngle[1] = StartEyeAngle[1];
      MovePosition = StartPosition;
      EyePosition = MovePosition;
      Mode = MODE_SHADED; /* also use default view and floor mode again */
      FloorMode = MODE_FLOOR_CHECKER;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_UP:
    case EVENT_UP2:
      MovePosition.z += ELEV_STEP;
      EyePosition = MovePosition;
      glutPostRedisplay();
      break;
    case EVENT_DOWN:
      MovePosition.z -= ELEV_STEP;
      EyePosition = MovePosition;
      glutPostRedisplay();
      break;
    case EVENT_FORWARD:
      MovePosition.x += o;
      MovePosition.y += a;
      EyePosition = MovePosition;
      glutPostRedisplay();
      break;
    case EVENT_BACK:
      MovePosition.x -= o;
      MovePosition.y -= a;
      EyePosition = MovePosition;
      glutPostRedisplay();
      break;
    case EVENT_LEFT:
      MovePosition.x -= a;
      MovePosition.y += o;
      EyePosition = MovePosition;
      glutPostRedisplay();
      break;
    case EVENT_RIGHT:
      MovePosition.x += a;
      MovePosition.y -= o;
      EyePosition = MovePosition;
      glutPostRedisplay();
      break;
    case EVENT_TURN_LEFT:
			EyeAngle[0] -= ROTATE_STEP;
      if (EyeAngle[0] < -180) {
        EyeAngle[0] += 360;
      }
      glutPostRedisplay();
      break;
    case EVENT_TURN_RIGHT:
			EyeAngle[0] += ROTATE_STEP;
      if (EyeAngle[0] > 180) {
        EyeAngle[0] -= 360;
      }
      glutPostRedisplay();
      break;
    case EVENT_REVERSE:
      EyeAngle[0] += 180;
      if (EyeAngle[0] > 180) {
        EyeAngle[0] -= 360;
      }
      glutPostRedisplay();
      break;
    case EVENT_ROT_UP:
			EyeAngle[1] -= ROTATE_STEP;
      if (EyeAngle[1] < -180) {
        EyeAngle[1] = -180;
      }
      glutPostRedisplay();
      break;
    case EVENT_ROT_DOWN:
    case EVENT_ROT_DOWN2:
			EyeAngle[1] += ROTATE_STEP;
      if (EyeAngle[1] > 0) {
        EyeAngle[1] = 0;
      }
      glutPostRedisplay();
      break;
    case EVENT_SHADED:
      Mode = MODE_SHADED;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_TRI:
      Mode = MODE_TRI;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_STRIP:
      Mode = MODE_STRIP;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_QUADS:
      Mode = MODE_QUADS;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_FLOOR_NO:
      FloorMode = MODE_FLOOR_NO;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_FLOOR_GREEN:
      FloorMode = MODE_FLOOR_GREEN;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
    case EVENT_FLOOR_CHECKER:
      FloorMode = MODE_FLOOR_CHECKER;
      ModeChange = TRUE;
      glutPostRedisplay();
      break;
  }
}


/* wrapper function to redirect menu events */
/* to the key processing function           */
void menuEvent(int menu) {
  keyboard(menu, 0, 0);
}


/* "inline" function for speed purposes:       */
/* calculates normal n given by triangle a,b,c */
#define TRIANGLE_NORMAL(a, b, c, n) \
  n = ((b)-(a))^((c)-(a)); \
  n.Normalize();


/* for each triangle formed by 3 points calculate normal vector;  */
/* 2 point from one scan line and the corresponding 2 points from */
/* the next scan line form 2 triangles, i.e. 2 normal vectors     */
/* belong to each point on each scan line but the last one        */
void calcNormals(void) {
  int num, c;

  /* process all scan lines but the last one */
  for (num=0; num<Num3d-1; num++) {
    /* process all scan points (includes the projected ones) but the last */
    for (c=0; c<NUM_SCANS+1; c++) {
      /* each new point defines two new triangles */
      TRIANGLE_NORMAL(Scan[num][c], Scan[num+1][c], Scan[num][c+1], Normal0[num][c]);
      TRIANGLE_NORMAL(Scan[num][c+1], Scan[num+1][c], Scan[num+1][c+1], Normal1[num][c]);
    }
  }
}



/***** global functions *****/


/* init OpenGL */
void grInitOpenGl(int *argcp, char** argv) {
  /* command line may contain OpenGL arguments */
  glutInit(argcp, argv);
  
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH);
}


/* set round or straight scan mode (information taken from command line) */
void grSetScanMode(int mode) {
  ScanMode = mode;
}


/* creates main window with title and pop-up menu */
void grCreateWindow(char *title) {  
  int menuFloor, menuMode;

  /* generate window with title */
  glutInitWindowSize(WindowWidth, WindowHeight); 
  glutInitWindowPosition(WindowPosX, WindowPosY);
  glutCreateWindow(title);

  /* allocate space for an OpenGL list */
  GlList = glGenLists(1);

  /* add pop-up menus */
  menuFloor = glutCreateMenu(menuEvent);
  glutAddMenuEntry("No", EVENT_FLOOR_NO);
  glutAddMenuEntry("Green", EVENT_FLOOR_GREEN);
  glutAddMenuEntry("Checkerboard", EVENT_FLOOR_CHECKER);
  menuMode = glutCreateMenu(menuEvent);
  glutAddMenuEntry("Shaded triangles", EVENT_SHADED);
  glutAddMenuEntry("Triangles", EVENT_TRI);
  glutAddMenuEntry("Scan Stripes", EVENT_STRIP);
  glutAddMenuEntry("Quads", EVENT_QUADS);
  glutCreateMenu(menuEvent);
  glutAddSubMenu("Display mode", menuMode);
  glutAddSubMenu("Floor mode", menuFloor);
  glutAddMenuEntry("Reset View", EVENT_RESET);
  glutAddMenuEntry("Exit", EVENT_EXIT);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
  
}


/* registers event handling functions and calls */
/* the OpenGL main loop                         */
void grMainLoop(void) {
  calcNormals();
  
  /* register callback functions */
  glutDisplayFunc(drawScene); 
  glutKeyboardFunc(keyboard);

  changeFrustum();
  glutMainLoop(); /* should never return */
}


/* allocates storage for 3D Cartesian data and normal vectors */
void grAllocateMemory(int num) {

  /* allocate pointer space for arrays */
  Scan = (Vertex **)malloc(num*sizeof(Vertex *));
  Normal0 = (Vertex **)malloc(num*sizeof(Vertex *));
  Normal1 = (Vertex **)malloc(num*sizeof(Vertex *));
  if ((Scan == NULL) || (Normal0 == NULL) || (Normal1 == NULL)) {
    printf("ERROR: Couldn't allocate memory\n");
    exit3d(1);
  }

  /* allocate each element in each array */
  for (int c=0; c<num; c++) {
    Scan[c] = (Vertex *)malloc((NUM_SCANS+2)*sizeof(Vertex));
    Normal0[c] = (Vertex *)malloc((NUM_SCANS+2)*sizeof(Vertex));
    Normal1[c] = (Vertex *)malloc((NUM_SCANS+2)*sizeof(Vertex));
    if ((Scan[c] == NULL) || (Normal0[c] == NULL) || (Normal1[c] == NULL)) {
      printf("ERROR: Couldn't allocate memory\n");
      exit3d(1);
    }
  }

  Num3d = num;
}


/* frees the memory allocated with grAllocateMemory() */
void grFree(void) {
  if ((Scan != NULL) && (Normal0 != NULL) && (Normal1 != NULL)) {
    for (int c=0; c<Num3d; c++) {
      free(Scan[c]);
      free(Normal0[c]);
      free(Normal1[c]);
    }
    free(Scan);
    free(Normal0);
    free(Normal1);
  }
}
