/*
 * laser3d - 3D visualization of laser data acquired by robot; enables walk through
 *
 * Filename    : laser3d.c
 * Author      : Michael Kaess
 * Initial     : 02/15/2001
 * Last Change : 12/27/2001
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>   /* getopt() */

#include "graphics.h"
#include "data.h"

#include "laser3d.h"



/***** global variables *****/


/* stores data from command line options */
char *ProgName;
char *DataFileName;
char *WriteDataFileName = NULL;
char *PgmReadName = NULL;
char *PgmWriteName = NULL;
int Interpolate = FALSE;

/* print information about processed data? */
int Verbose = TRUE;

/* for getopt() */
extern int optind;
extern char *optarg;


/***** local functions *****/


/* frees all allocated memory and exits returning error code err */
void exit3d(int err) {
  datFree();
  grFree();
  
  exit(err); /* the only exit in this program */
}


/* prints graphics window usage information;          */
/* printed to terminal when graphics window is opened */
void print3dKeys(void) {
  printf("KEYS for OpenGL Window:\n");
  printf("  General:\n");
  printf("    R  Reset view\n");
  printf("    Q  Quit program\n");
  printf("  Navigation:\n");
  printf("    y  Move up\n");
  printf("    h  Move down\n");
  printf("    i  Move forward\n");
  printf("    k  Move backward\n");
  printf("    u  Move left\n");
  printf("    o  Move right\n");
  printf("    j  Turn left\n");
  printf("    l  Turn right\n");
  printf("    r  Reverse (turn 180 degrees)\n");
  printf("    p  Rotate up\n");
  printf("    ;  Rotate down\n");
  printf("  Floor Mode:\n");
  printf("    a  No Floor\n");
  printf("    s  Green Floor\n");
  printf("    d  Checkerboard Floor (default)\n");
  printf("  Display Mode:\n");
  printf("    1  Shaded mode (default)\n");
  printf("    2  Triangle mode (random colors for each triangle)\n");
  printf("    3  Quad strip mode (random colors for each strip)\n");
  printf("    4  Quad mode (random colors for each quad)\n");
}


/* prints command line options; printed when    */
/* encountering wrong command line option or -h */
void printUsage(void) {
  printf("USAGE: %s -h | -H | [<OpenGl-opts>] [-vqmcrsCI8]\n", ProgName);
  printf("               [-O <output-file> | -S <file.pgm>]\n");
  printf("               {-R <file.pgm> | <input-file>}\n");
  printf("  OPTIONS:\n");
  printf("  -h  Print this help message and exit\n");
  printf("  -H  Print 3D window keys and exit\n");
  printf("  -v  verbose mode (default)\n");
  printf("  -q  quiet mode\n");
  printf("  -m  mm resolution (default)\n");
  printf("  -c  cm resolution\n");
  printf("  -s  straight scan (default)\n");
  printf("  -r  round scan\n");
  printf("  -C  special clean up for 360 degree round scans only\n");
  printf("  -I  interpolate over laser errors\n");
  printf("  -8  write .pgm in 8 bit format (default is 16)\n");
  printf("  -O <output-file>   write cleaned data to laser data format file\n");
  printf("  -S <file.pgm>      write raw data to pgm format file\n");
  printf("  -R <file.pgm>      read raw data from pgm format file\n");
  printf("  <input-file>       file containing the raw laser data as created\n");
  printf("                     by laserfit (hserver/MissionLab)\n");
}


/* reads in all information from the command line */
/* and sets the corresponding flags / strings     */
void processCommandLine(int argc, char** argv) {
  int c;
  int add; /* number of additional arguments necessary */
  
  ProgName = argv[0];

  while ((c = getopt(argc, argv, "mcrshvqH8CIO:S:R:") ) != -1) {
    /* each option character has to be in the string in getopt(); */
    /* a colon after an option means that there is an extra       */
    /* parameter to this option; 'W' is a reserved character      */
    switch (c) {
      case 'm':
        /* laser scanner was set to millimeter mode */
        datSetResolution(RESOLUTION_MM);
        break;
      case 'c':
        /* laser scanner was set to centimeter mode */
        datSetResolution(RESOLUTION_CM);
        break;
      case 'r':
        /* robot rotated instead of moving */
        grSetScanMode(ROUND_SCAN);
        break;
      case 's':
        /* robot moved in a straight line */
        grSetScanMode(STRAIGHT_SCAN);
        break;
      case '8':
        /* write pgm file in 8-bit mode; helpful for viewing  */
        /* these images, 16-bit pgms are not widely supported */
        datSetPgmMode(PGM_8_BIT);
        break;
      case 'C':
        /* for 360 degree scans only: remove all scans but */
        /* the data from one complete 360 degree turn      */
        datSet360Mode();
        break;
      case 'I':
        /* interpolate over laser scanner errors */
        Interpolate = TRUE;
        break;
      case 'O':
        /* write cleaned data to laser data file */
        WriteDataFileName = optarg;
        break;
      case 'R':
        /* read data from pgm file instead of laser data file */
        PgmReadName = optarg;
        break;
      case 'S':
        /* save to pgm file instead of displaying the data */
        PgmWriteName = optarg;
        break;
      case 'v':
        /* show informations and warning messages */
        Verbose = TRUE;
        break;
      case 'q':
        /* allow only error messages */
        Verbose = FALSE;
        break;
      case 'H':
        print3dKeys();
        exit3d(0);
        break;
      case 'h':
        printUsage();
        exit3d(0);
        break;
      case '?': /* '?' means unknown option, from getopt() */
        printUsage();
        exit3d(1);
        break;
    }
  }

  /* if not pgm read mode (-R): laser data file name necessary */
  add = 0;
  if ((PgmReadName == NULL) || ((PgmReadName != NULL) && (PgmWriteName == NULL))) {
    add = 1;
  }
  
  if (argc > optind+add) {
    printf("ERROR: Too many arguments.\n");
    printUsage();
    exit3d(1);
  } else if (argc < optind+add) {
    printf("ERROR: Input file name missing.\n");
    printUsage();
    exit3d(1);
  }

  if (add == 1) {
    /* assume that additional argument is the laser data file name */
    DataFileName = argv[optind];
  }
  
}


/**************************************************/
int main(int argc, char** argv) {

  char title[200]; /* graphics window title */
  int start, end;  /* defines range of laser scans displayed/processed */

  /* has to be first, may modify argc and *argv */
  grInitOpenGl(&argc, argv);
  
  processCommandLine(argc, argv);

  start = 0;

  if (DataFileName != NULL) {
    /* read in laser data file; */
    /* sets [start, end) to a range of valid data */
    datGetRaw3dData(DataFileName, &start, &end);
  }

  if (PgmReadName != NULL) {
    /* read in raw pgm image file */
    end = datReadRawPgmFile(PgmReadName, &start);
  }

  if (Interpolate == TRUE) {
    datInterpolateData(start, end);
  }
  
  if (WriteDataFileName != NULL) {
    /* write stripped and corrected data to file using the */
    /* same laser data format as generated by laserfit     */
    datWriteCleanedFile(WriteDataFileName, start, end);
  } else if (PgmWriteName != NULL) {
    /* write raw data to raw pgm image file */
    datWriteRawPgmFile(PgmWriteName, start, end);
  } else {
    /* display 3D model */
    
    /* allocate the 3D data structure */
    grAllocateMemory(end-start+1);
    /* convert data from raw to 3D */
    datRawTo3d(start, end);

    /* create window title and open window */
    strcpy(title, ProgName);
    strcat(title, ": ");
    if (DataFileName != NULL) {
      strcat(title, DataFileName);
    } else if (PgmReadName != NULL) {
      strcat(title, PgmReadName);
    }
    grCreateWindow(title);

    /* print usage of graphics window if not in quiet mode */
    if (Verbose == TRUE) {
      print3dKeys();
    }
    
    /* should never return */
    grMainLoop();
  }

  exit3d(0);
}
