/*
 * laser3d - 3D visualization of laser data acquired by robot; enables walk through
 *
 * Filename    : laser3d.h
 * Author      : Michael Kaess
 * Initial     : 02/15/2001
 * Last Change : 12/27/2001
 *
 */


#ifndef _LASER3D_INC
#define _LASER3D_INC

#include <sys/time.h>



/* usually laserfit stores the angle of the robot in degrees;    */
/* this option makes the program convert from degrees to radians */
#define DATAFILE_USES_DEGREES


/***** ADAPT THE FOLLOWING CONSTANTS TO YOUR EXPERIMENT SETUP *****/


/* laser scanner mount location */
#define SCAN_HEIGHT_ROUND      0.965 /* height of center of laser scanner in meter */
#define SCAN_RADIUS_ROUND    210     /* radius the center of the laser rotated with in mm */
#define SCAN_TILT_ROUND        1.9   /* in degrees, laser was tilted because casing is not flat... */
#define SCAN_HEIGHT_STRAIGHT   0.52  /* height of center of laser in meter */

/* laser scanner properties */
#define SCALE_LASER_MM  0.001  /* mm resolution: 0.001m */
#define SCALE_LASER_CM  0.01   /* cm resolution: 0.01m */
#define NUM_SCANS     361      /* 180, 0.5 degree increments */

/* which values should be considered invalid? */
/* our SICK LMS200 usually returns 8187 as invalid... */
#define LASER_MAX_VALUE_MM  8187  /* removes far points / wrong measurements  */
#define LASER_MAX_VALUE_CM  8100  /* ...   >= 8192 removes no points, 12bit ! */


/***** ADAPT THE FOLLOWING CONSTANTS TO YOUR VIRTUAL MODEL PREFERENCES *****/


/* 3D window setup */
#define WINDOW_WIDTH   720 /* initial window size */
#define WINDOW_HEIGHT  480
#define WINDOW_POS_X   100 /* initial window position */
#define WINDOW_POS_Y   100

/* navigation, virtual dimensions */
#define EYE_HEIGHT   1.7f  /* initial height of viewpoint in meter */
#define MOVE_STEP    0.1   /* distance to move forward, back, right or left */
#define ELEV_STEP    0.1   /* distance to move up or down */
#define ROTATE_STEP  6     /* angle of rotation step */

/* dimensions and position of model */
#define SCALE 1
#define SCALE_X SCALE
#define SCALE_Y SCALE
#define SCALE_Z SCALE
#define OFFSET_X 0
#define OFFSET_Y 0
#define OFFSET_Z 0

/* checker board dimensions */
#define CB_GRID_NUM  40   /* number of quads in each, x and y direction */
#define CB_QUAD_SIZE  0.5 /* length of one side of each quad in meters */



/***** NO NEED TO CHANGE ANYTHING BELOW THIS LINE *****/



/* view modes */
#define MODE_SHADED  1 /* same color, using light */
#define MODE_TRI     2 /* triangles with random colors */
#define MODE_STRIP   3 /* quad stripes with random colors */
#define MODE_QUADS   4 /* quads with random colors */
#define MODE_FLOOR_NO       0 /* nothing */
#define MODE_FLOOR_GREEN    1 /* green */
#define MODE_FLOOR_CHECKER  2 /* checker board */

/* key mapping */
#define EVENT_FLOOR_NO      'a'
#define EVENT_FLOOR_GREEN   's'
#define EVENT_FLOOR_CHECKER 'd'
#define EVENT_SHADED        '1'
#define EVENT_TRI           '2'
#define EVENT_STRIP         '3'
#define EVENT_QUADS         '4'
#define EVENT_RESET         'R'
#define EVENT_EXIT          'Q'
#define EVENT_EXIT2          27
#define EVENT_UP            'y'
#define EVENT_UP2           'z' /* german kb */
#define EVENT_DOWN          'h'
#define EVENT_FORWARD       'i'
#define EVENT_BACK          'k'
#define EVENT_LEFT          'u'
#define EVENT_RIGHT         'o'
#define EVENT_TURN_LEFT     'j'
#define EVENT_TURN_RIGHT    'l'
#define EVENT_REVERSE       'r'
#define EVENT_ROT_UP        'p'
#define EVENT_ROT_DOWN      ';'
#define EVENT_ROT_DOWN2     246 /* german kb */


/* do not change all following definitions */
#define INVALID  -10000000  /* invalid float value */
#define RESOLUTION_CM  0
#define RESOLUTION_MM  1
#define TRUE   1
#define FALSE  0
#define ROUND_SCAN     0
#define STRAIGHT_SCAN  1
#define PGM_8_BIT  0
#define PGM_16_BIT 1


/* has to be accessed by all modules to allow shutting */
/* down the program after an error occured             */
void exit3d(int err);


/***** types *****/


/* raw laser scan data */
typedef struct SRawScan {
  unsigned int dist[NUM_SCANS]; /* measured distances of one 180 scan */
  float x, y, t;                /* pose of robot */
  struct timeval tstamp;        /* time data was acquired */
} __xx; /* some compilers require a name... */
typedef struct SRawScan TRawScan;
typedef TRawScan *PRawScan;


#endif
